/*
 * tangle-scrolling-actor.c
 *
 * This file is part of Tangle Toolkit - A graphical actor library based on Clutter Toolkit
 *
 * (c) 2010 Henrik Hedberg <henrik.hedberg@innologies.fi>
 *
 */

#include "tangle-scrolling-actor.h"
#include "tangle-clone.h"
#include "tangle-vault.h"

/**
 * SECTION:tangle-scrolling-actor
 * @Short_description: A wrapper actor that scrolls its content
 * @Title: TangleScrollingActor
 */

G_DEFINE_TYPE(TangleScrollingActor, tangle_scrolling_actor, TANGLE_TYPE_WRAPPER_ACTOR);

enum {
	PROP_0,
	PROP_SCROLLING_OFFSET_X,
	PROP_SCROLLING_OFFSET_Y,
	PROP_SCROLLING_WIDTH,
	PROP_SCROLLING_HEIGHT,
	PROP_SCROLLING_THRESHOLD_X,
	PROP_SCROLLING_THRESHOLD_Y,
	PROP_MAX_OVERSHOOT_X,
	PROP_MAX_OVERSHOOT_Y,
	PROP_REPEAT_X,
	PROP_REPEAT_Y,
	PROP_CONTINUOUS_X,
	PROP_CONTINUOUS_Y,
	PROP_KINETIC_SCROLLING,
	PROP_SCROLLING_DECELERATION
};

enum {
	CLAMP_SCROLLING_OFFSET_X,
	CLAMP_SCROLLING_OFFSET_Y,
	LAST_SIGNAL
};

typedef enum {
	HORIZONTAL = 0,
	VERTICAL
} Axis;

struct _TangleScrollingActorPrivate {
	gfloat scrolling_offset[2];
	gfloat scrolling_threshold[2];
	gfloat max_overshoot[2];
	gdouble scrolling_deceleration;
	gboolean repeat[2];
	gboolean continuous[2];

	gfloat scrolling_length[2];
	
	gboolean scrolling_effective[2];

	gfloat start_motion[2];
	guint32 start_motion_time[2];
	gfloat farest_motion[2];
	guint32 farest_motion_time[2];
	
	guint32 previous_motion_time;

	gfloat start_scrolling_motion[2];
	gfloat start_scrolling_offset[2];

	gboolean going_forward[2];

	gulong captured_event_handler_id;
	ClutterAnimation* animation[2];
	
	ClutterActor* clones[8];

	guint kinetic_scrolling : 1;

	guint descendant_interacting : 1;	
	guint descendant_dragging : 1;
};

static guint signals[LAST_SIGNAL] = { 0 };

static gboolean on_captured_event(ClutterActor* actor, ClutterEvent* event, gpointer user_data);
static gfloat get_length(TangleScrollingActor* scrolling_actor, Axis axis);
static gfloat normalise_scrolling_offset(TangleScrollingActor* scrolling_actor, Axis axis, gfloat scrolling_offset, gfloat* delta_return);
static void stop_interacting(TangleScrollingActor* scrolling_actor);
static void setup_clones(TangleScrollingActor* scrolling_actor);

ClutterActor* tangle_scrolling_actor_new(ClutterActor* wrapped) {

	return CLUTTER_ACTOR(g_object_new(TANGLE_TYPE_SCROLLING_ACTOR, "wrapped", wrapped, NULL));
}

gfloat tangle_scrolling_actor_get_scrolling_offset_x(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_offset[HORIZONTAL];
}

void tangle_scrolling_actor_set_scrolling_offset_x(TangleScrollingActor* scrolling_actor, gfloat scrolling_offset_x) {
	gfloat width;
	
	if (scrolling_actor->priv->scrolling_offset[HORIZONTAL] != scrolling_offset_x) {
		scrolling_actor->priv->scrolling_offset[HORIZONTAL] = normalise_scrolling_offset(scrolling_actor, HORIZONTAL, scrolling_offset_x, NULL);
		clutter_actor_queue_relayout(CLUTTER_ACTOR(scrolling_actor));
		g_object_notify(G_OBJECT(scrolling_actor), "scrolling-offset-x");
	}	
}

gfloat tangle_scrolling_actor_get_scrolling_offset_y(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_offset[VERTICAL];
}

void tangle_scrolling_actor_set_scrolling_offset_y(TangleScrollingActor* scrolling_actor, gfloat scrolling_offset_y) {
	gfloat height;
	
	if (scrolling_actor->priv->scrolling_offset[VERTICAL] != scrolling_offset_y) {
		scrolling_actor->priv->scrolling_offset[VERTICAL] = normalise_scrolling_offset(scrolling_actor, VERTICAL, scrolling_offset_y, NULL);
		clutter_actor_queue_relayout(CLUTTER_ACTOR(scrolling_actor));
		g_object_notify(G_OBJECT(scrolling_actor), "scrolling-offset-y");
	}
}

gfloat tangle_scrolling_actor_get_scrolling_width(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_length[HORIZONTAL];
}

gfloat tangle_scrolling_actor_get_scrolling_height(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_length[VERTICAL];
}

void tangle_scrolling_actor_set_scrolling_threshold_x(TangleScrollingActor* scrolling_actor, gfloat scrolling_threshold_x) {
	if (scrolling_actor->priv->scrolling_threshold[HORIZONTAL] != scrolling_threshold_x) {
		scrolling_actor->priv->scrolling_threshold[HORIZONTAL] = scrolling_threshold_x;
		g_object_notify(G_OBJECT(scrolling_actor), "scrolling-threshold-x");
	}
}

gfloat tangle_scrolling_actor_get_scrolling_threshold_y(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_threshold[VERTICAL];
}

void tangle_scrolling_actor_set_scrolling_threshold_y(TangleScrollingActor* scrolling_actor, gfloat scrolling_threshold_y) {
	if (scrolling_actor->priv->scrolling_threshold[VERTICAL] != scrolling_threshold_y) {
		scrolling_actor->priv->scrolling_threshold[VERTICAL] = scrolling_threshold_y;
		g_object_notify(G_OBJECT(scrolling_actor), "scrolling-threshold-y");
	}
}

void tangle_scrolling_actor_set_scrolling_threshold(TangleScrollingActor* scrolling_actor, gfloat scrolling_threshold) {
	tangle_scrolling_actor_set_scrolling_threshold_x(scrolling_actor, scrolling_threshold);
	tangle_scrolling_actor_set_scrolling_threshold_y(scrolling_actor, scrolling_threshold);
}

gboolean tangle_scrolling_actor_get_repeat_x(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->repeat[HORIZONTAL];
}

void tangle_scrolling_actor_set_repeat_x(TangleScrollingActor* scrolling_actor, gboolean repeat_x) {
	if (scrolling_actor->priv->repeat[HORIZONTAL] != repeat_x) {
		g_object_freeze_notify(G_OBJECT(scrolling_actor));

		scrolling_actor->priv->repeat[HORIZONTAL] = repeat_x;
		g_object_notify(G_OBJECT(scrolling_actor), "repeat-x");
		
		if (!repeat_x) {
			tangle_scrolling_actor_set_continuous_x(scrolling_actor, FALSE);
		}

		g_object_thaw_notify(G_OBJECT(scrolling_actor));
	}
}

gboolean tangle_scrolling_actor_get_repeat_y(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->repeat[VERTICAL];
}

void tangle_scrolling_actor_set_repeat_y(TangleScrollingActor* scrolling_actor, gboolean repeat_y) {
	if (scrolling_actor->priv->repeat[VERTICAL] != repeat_y) {
		g_object_freeze_notify(G_OBJECT(scrolling_actor));

		scrolling_actor->priv->repeat[VERTICAL] = repeat_y;
		g_object_notify(G_OBJECT(scrolling_actor), "repeat-y");
		
		if (!repeat_y) {
			tangle_scrolling_actor_set_continuous_y(scrolling_actor, FALSE);
		}

		g_object_thaw_notify(G_OBJECT(scrolling_actor));
	}
}

gboolean tangle_scrolling_actor_get_continuous_x(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->continuous[HORIZONTAL];
}

void tangle_scrolling_actor_set_continuous_x(TangleScrollingActor* scrolling_actor, gboolean continuous_x) {
	if (scrolling_actor->priv->continuous[HORIZONTAL] != continuous_x) {
		g_object_freeze_notify(G_OBJECT(scrolling_actor));

		scrolling_actor->priv->continuous[HORIZONTAL] = continuous_x;
		g_object_notify(G_OBJECT(scrolling_actor), "continuous-x");

		if (continuous_x) {
			tangle_scrolling_actor_set_repeat_x(scrolling_actor, TRUE);
		}
		
		setup_clones(scrolling_actor);

		g_object_thaw_notify(G_OBJECT(scrolling_actor));
	}
}

gboolean tangle_scrolling_actor_get_continuous_y(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->continuous[VERTICAL];
}

void tangle_scrolling_actor_set_continuous_y(TangleScrollingActor* scrolling_actor, gboolean continuous_y) {
	if (scrolling_actor->priv->continuous[VERTICAL] != continuous_y) {
		g_object_freeze_notify(G_OBJECT(scrolling_actor));

		scrolling_actor->priv->continuous[VERTICAL] = continuous_y;
		g_object_notify(G_OBJECT(scrolling_actor), "continuous-y");
		
		if (continuous_y) {
			tangle_scrolling_actor_set_repeat_y(scrolling_actor, TRUE);
		}

		setup_clones(scrolling_actor);

		g_object_thaw_notify(G_OBJECT(scrolling_actor));
	}
}

gboolean tangle_scrolling_actor_get_kinetic_scrolling(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->kinetic_scrolling;
}

void tangle_scrolling_actor_set_kinetic_scrolling(TangleScrollingActor* scrolling_actor, gboolean kinetic_scrolling) {
	scrolling_actor->priv->kinetic_scrolling = kinetic_scrolling;
}

gdouble tangle_scrolling_actor_get_scrolling_deceleration(TangleScrollingActor* scrolling_actor) {

	return scrolling_actor->priv->scrolling_deceleration;
}

void tangle_scrolling_actor_set_scrolling_deceleration(TangleScrollingActor* scrolling_actor, gdouble scrolling_deceleration) {
	scrolling_actor->priv->scrolling_deceleration = scrolling_deceleration;
}

void tangle_scrolling_actor_clamp_scrolling_offset_x(TangleScrollingActor* scrolling_actor) {
	TangleClamp* clamp;
	
	clamp = tangle_clamp_new(TANGLE_CLAMP_HORIZONTAL, scrolling_actor->priv->scrolling_offset[HORIZONTAL]);
	g_signal_emit(scrolling_actor, signals[CLAMP_SCROLLING_OFFSET_X], 0, clamp);
	if (tangle_clamp_get_clamped_value_set(clamp)) {
		tangle_actor_animate(TANGLE_ACTOR(scrolling_actor), CLUTTER_EASE_IN_OUT_QUAD, 300, "scrolling-offset-x", tangle_clamp_get_clamped_value(clamp), NULL);		
	}
	g_object_unref(clamp);
}

void tangle_scrolling_actor_clamp_scrolling_offset_y(TangleScrollingActor* scrolling_actor) {
	TangleClamp* clamp;
	
	clamp = tangle_clamp_new(TANGLE_CLAMP_VERTICAL, scrolling_actor->priv->scrolling_offset[VERTICAL]);
	g_signal_emit(scrolling_actor, signals[CLAMP_SCROLLING_OFFSET_Y], 0, clamp);
	if (tangle_clamp_get_clamped_value_set(clamp)) {
		tangle_actor_animate(TANGLE_ACTOR(scrolling_actor), CLUTTER_EASE_IN_OUT_QUAD, 300, "scrolling-offset-y", tangle_clamp_get_clamped_value(clamp), NULL);		
	}
	g_object_unref(clamp);
}

void tangle_scrolling_actor_clamp_offsets(TangleScrollingActor* scrolling_actor) {
	tangle_scrolling_actor_clamp_scrolling_offset_x(scrolling_actor);
	tangle_scrolling_actor_clamp_scrolling_offset_y(scrolling_actor);
}

void tangle_scrolling_actor_clamp_page_boundaries(TangleScrollingActor* scrolling_actor, TangleClamp* clamp) {
	TangleClampDirection direction;
	gfloat original_value;
	gfloat page_size;
	
	g_return_if_fail(TANGLE_IS_SCROLLING_ACTOR(scrolling_actor));
	
	direction = tangle_clamp_get_direction(clamp);
	original_value = tangle_clamp_get_original_value(clamp);
	
	if (direction == TANGLE_CLAMP_HORIZONTAL) {
		if (original_value < 0.0) {
			tangle_clamp_set_clamped_value(clamp, 0.0);
		} else if (original_value > scrolling_actor->priv->scrolling_length[HORIZONTAL]) {
			tangle_clamp_set_clamped_value(clamp, scrolling_actor->priv->scrolling_length[HORIZONTAL]);
		} else {
			page_size = get_length(scrolling_actor, HORIZONTAL);
			tangle_clamp_set_clamped_value(clamp, ((guint)((original_value + page_size / 2) / page_size)) * page_size);
		}
	} else if (direction == TANGLE_CLAMP_VERTICAL) {
		if (original_value < 0.0) {
			tangle_clamp_set_clamped_value(clamp, 0.0);
		} else if (original_value > scrolling_actor->priv->scrolling_length[VERTICAL]) {
			tangle_clamp_set_clamped_value(clamp, scrolling_actor->priv->scrolling_length[VERTICAL]);
		} else {
			page_size = get_length(scrolling_actor, VERTICAL);
			tangle_clamp_set_clamped_value(clamp, ((guint)((original_value + page_size / 2) / page_size)) * page_size);
		}
	}
}

static gboolean tangle_scrolling_actor_button_press_event(ClutterActor* actor, ClutterButtonEvent* event) {
	TangleScrollingActor* scrolling_actor;
		
	scrolling_actor = TANGLE_SCROLLING_ACTOR(actor);

	if (!scrolling_actor->priv->captured_event_handler_id) {
		if (scrolling_actor->priv->animation[HORIZONTAL]) {
			tangle_actor_stop_animation(TANGLE_ACTOR(scrolling_actor), scrolling_actor->priv->animation[HORIZONTAL]);
		}
		if (scrolling_actor->priv->animation[VERTICAL]) {
			tangle_actor_stop_animation(TANGLE_ACTOR(scrolling_actor), scrolling_actor->priv->animation[VERTICAL]);
		}
		scrolling_actor->priv->start_scrolling_motion[HORIZONTAL] = event->x;
		scrolling_actor->priv->start_scrolling_motion[VERTICAL] = event->y;
		scrolling_actor->priv->scrolling_effective[HORIZONTAL] = scrolling_actor->priv->scrolling_effective[VERTICAL] = FALSE;
		scrolling_actor->priv->captured_event_handler_id = g_signal_connect(clutter_actor_get_stage(actor), "captured-event", G_CALLBACK(on_captured_event), scrolling_actor);
	}

	return FALSE;
}


static void tangle_scrolling_actor_allocate_wrapped(TangleWrapperActor* wrapper_actor, ClutterActor* wrapped, const ClutterActorBox* box, ClutterAllocationFlags flags) {
	TangleScrollingActor* scrolling_actor;
	gfloat old_width, old_height;
	ClutterActorBox actor_box;

	scrolling_actor = TANGLE_SCROLLING_ACTOR(wrapper_actor);

	old_width = scrolling_actor->priv->scrolling_length[HORIZONTAL];
	old_height = scrolling_actor->priv->scrolling_length[VERTICAL];
	clutter_actor_get_preferred_size(wrapped, NULL, NULL, &scrolling_actor->priv->scrolling_length[HORIZONTAL], &scrolling_actor->priv->scrolling_length[VERTICAL]);
	if (scrolling_actor->priv->scrolling_length[HORIZONTAL] < box->x2 - box->x1) {
		scrolling_actor->priv->scrolling_length[HORIZONTAL] = box->x2 - box->x1;
	}
	if (scrolling_actor->priv->scrolling_length[VERTICAL] < box->y2 - box->y1) {
		scrolling_actor->priv->scrolling_length[VERTICAL] = box->y2 - box->y1;
	}
	if (old_width != scrolling_actor->priv->scrolling_length[HORIZONTAL]) {
		g_object_notify(G_OBJECT(wrapper_actor), "scrolling-width");
	}
	if (old_height != scrolling_actor->priv->scrolling_length[VERTICAL]) {
		g_object_notify(G_OBJECT(wrapper_actor), "scrolling-height");
	}

	actor_box.x1 = box->x1 - scrolling_actor->priv->scrolling_offset[HORIZONTAL];
	actor_box.y1 = box->y1 - scrolling_actor->priv->scrolling_offset[VERTICAL];
	actor_box.x2 = actor_box.x1 + scrolling_actor->priv->scrolling_length[HORIZONTAL];
	actor_box.y2 = actor_box.y1 + scrolling_actor->priv->scrolling_length[VERTICAL];

	clutter_actor_allocate(wrapped, &actor_box, flags);
	
	/* 0 1 2
	   3 X 4
	   5 6 7 */
	   
	if (scrolling_actor->priv->clones[0]) {
		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[0], &actor_box, flags);

		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];
	}
		
	if (scrolling_actor->priv->clones[1]) {
		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[1], &actor_box, flags);

		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];
	}

	if (scrolling_actor->priv->clones[2]) {
		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[2], &actor_box, flags);

		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];
	}

	if (scrolling_actor->priv->clones[3]) {
		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[3], &actor_box, flags);

		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
	}

	if (scrolling_actor->priv->clones[4]) {
		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[4], &actor_box, flags);

		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
	}

	if (scrolling_actor->priv->clones[5]) {
		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[5], &actor_box, flags);

		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];
	}

	if (scrolling_actor->priv->clones[6]) {
		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[6], &actor_box, flags);

		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];
	}

	if (scrolling_actor->priv->clones[7]) {
		actor_box.x1 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 += scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 += scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 += scrolling_actor->priv->scrolling_length[VERTICAL];

		clutter_actor_allocate(scrolling_actor->priv->clones[7], &actor_box, flags);

		actor_box.x1 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.x2 -= scrolling_actor->priv->scrolling_length[HORIZONTAL];
		actor_box.y1 -= scrolling_actor->priv->scrolling_length[VERTICAL];
		actor_box.y2 -= scrolling_actor->priv->scrolling_length[VERTICAL];
	}
	
	scrolling_actor->priv->scrolling_length[HORIZONTAL] -= box->x2 - box->x1;
	scrolling_actor->priv->scrolling_length[VERTICAL] -= box->y2 - box->y1;
}

static void tangle_scrolling_actor_paint_wrapped(TangleWrapperActor* wrapper_actor, ClutterActor* wrapped) {
	TangleScrollingActor* scrolling_actor;
	gint i;
	
	scrolling_actor = TANGLE_SCROLLING_ACTOR(wrapper_actor);

	TANGLE_WRAPPER_ACTOR_CLASS(tangle_scrolling_actor_parent_class)->paint_wrapped(wrapper_actor, wrapped);

	for (i = 0; i < 8; i++) {
		if (scrolling_actor->priv->clones[i]) {
			clutter_actor_paint(scrolling_actor->priv->clones[i]);
		}
	}
}

static void tangle_scrolling_actor_clamp_scrolling_offset_x_impl(TangleScrollingActor* scrolling_actor, TangleClamp* clamp) {
	if (!tangle_clamp_get_clamped_value_set(clamp) && !scrolling_actor->priv->continuous[HORIZONTAL]) {
		if (scrolling_actor->priv->scrolling_offset[HORIZONTAL] < 0.0) {
			tangle_clamp_set_clamped_value(clamp, 0.0);
		} else if (scrolling_actor->priv->scrolling_offset[HORIZONTAL] > scrolling_actor->priv->scrolling_length[HORIZONTAL]) {
			tangle_clamp_set_clamped_value(clamp, scrolling_actor->priv->scrolling_length[HORIZONTAL]);
		}
	}
}

static void tangle_scrolling_actor_clamp_scrolling_offset_y_impl(TangleScrollingActor* scrolling_actor, TangleClamp* clamp) {
	if (!tangle_clamp_get_clamped_value_set(clamp) && !scrolling_actor->priv->continuous[VERTICAL]) {
		if (scrolling_actor->priv->scrolling_offset[VERTICAL] < 0.0) {
			tangle_clamp_set_clamped_value(clamp, 0.0);
		} else if (scrolling_actor->priv->scrolling_offset[VERTICAL] > scrolling_actor->priv->scrolling_length[VERTICAL]) {
			tangle_clamp_set_clamped_value(clamp, scrolling_actor->priv->scrolling_length[VERTICAL]);
		}
	}
}

static void tangle_scrolling_actor_map(ClutterActor* actor) {
	TangleScrollingActor* scrolling_actor;
	gint i;
	
	scrolling_actor = TANGLE_SCROLLING_ACTOR(actor);
	
	CLUTTER_ACTOR_CLASS(tangle_scrolling_actor_parent_class)->map(actor);

	for (i = 0; i < 8; i++) {
		if (scrolling_actor->priv->clones[i]) {
			clutter_actor_map(scrolling_actor->priv->clones[i]);
		}
	}
}

static void tangle_scrolling_actor_set_property(GObject* object, guint prop_id, const GValue* value, GParamSpec* pspec) {
	TangleScrollingActor* scrolling_actor;
	
	scrolling_actor = TANGLE_SCROLLING_ACTOR(object);

	switch (prop_id) {
		case PROP_SCROLLING_OFFSET_X:
			tangle_scrolling_actor_set_scrolling_offset_x(scrolling_actor, g_value_get_float(value));
			break;
		case PROP_SCROLLING_OFFSET_Y:
			tangle_scrolling_actor_set_scrolling_offset_y(scrolling_actor, g_value_get_float(value));
			break;
		case PROP_SCROLLING_THRESHOLD_X:
			tangle_scrolling_actor_set_scrolling_threshold_x(scrolling_actor, g_value_get_float(value));
			break;
		case PROP_SCROLLING_THRESHOLD_Y:
			tangle_scrolling_actor_set_scrolling_threshold_y(scrolling_actor, g_value_get_float(value));
			break;
		case PROP_MAX_OVERSHOOT_X:
			scrolling_actor->priv->max_overshoot[HORIZONTAL] = g_value_get_float(value);
			break;
		case PROP_MAX_OVERSHOOT_Y:
			scrolling_actor->priv->max_overshoot[VERTICAL] = g_value_get_float(value);
			break;
		case PROP_REPEAT_X:
			tangle_scrolling_actor_set_repeat_x(scrolling_actor, g_value_get_boolean(value));
			break;
		case PROP_REPEAT_Y:
			tangle_scrolling_actor_set_repeat_y(scrolling_actor, g_value_get_boolean(value));
			break;
		case PROP_CONTINUOUS_X:
			tangle_scrolling_actor_set_continuous_x(scrolling_actor, g_value_get_boolean(value));
			break;
		case PROP_CONTINUOUS_Y:
			tangle_scrolling_actor_set_continuous_y(scrolling_actor, g_value_get_boolean(value));
			break;
		case PROP_KINETIC_SCROLLING:
			tangle_scrolling_actor_set_kinetic_scrolling(scrolling_actor, g_value_get_boolean(value));
			break;
		case PROP_SCROLLING_DECELERATION:
			tangle_scrolling_actor_set_scrolling_deceleration(scrolling_actor, g_value_get_double(value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
			break;
	}
}

static void tangle_scrolling_actor_get_property(GObject* object, guint prop_id, GValue* value, GParamSpec* pspec) {
        TangleScrollingActor* scrolling_actor;

	scrolling_actor = TANGLE_SCROLLING_ACTOR(object);

        switch (prop_id) {
		case PROP_SCROLLING_OFFSET_X:
			g_value_set_float(value, scrolling_actor->priv->scrolling_offset[HORIZONTAL]);
			break;
		case PROP_SCROLLING_OFFSET_Y:
			g_value_set_float(value, scrolling_actor->priv->scrolling_offset[VERTICAL]);
			break;
		case PROP_SCROLLING_WIDTH:
			g_value_set_float(value, scrolling_actor->priv->scrolling_length[HORIZONTAL]);
			break;
		case PROP_SCROLLING_HEIGHT:
			g_value_set_float(value, scrolling_actor->priv->scrolling_length[VERTICAL]);
			break;
		case PROP_SCROLLING_THRESHOLD_X:
			g_value_set_float(value, scrolling_actor->priv->scrolling_threshold[HORIZONTAL]);
			break;
		case PROP_SCROLLING_THRESHOLD_Y:
			g_value_set_float(value, scrolling_actor->priv->scrolling_threshold[VERTICAL]);
			break;
		case PROP_MAX_OVERSHOOT_X:
			g_value_set_float(value, scrolling_actor->priv->max_overshoot[HORIZONTAL]);
			break;
		case PROP_MAX_OVERSHOOT_Y:
			g_value_set_float(value, scrolling_actor->priv->max_overshoot[VERTICAL]);
			break;
		case PROP_REPEAT_X:
			g_value_set_boolean(value, scrolling_actor->priv->repeat[HORIZONTAL]);
			break;
		case PROP_REPEAT_Y:
			g_value_set_boolean(value, scrolling_actor->priv->repeat[VERTICAL]);
			break;
		case PROP_CONTINUOUS_X:
			g_value_set_boolean(value, scrolling_actor->priv->continuous[HORIZONTAL]);
			break;
		case PROP_CONTINUOUS_Y:
			g_value_set_boolean(value, scrolling_actor->priv->continuous[VERTICAL]);
			break;
		case PROP_KINETIC_SCROLLING:
			g_value_set_boolean(value, scrolling_actor->priv->kinetic_scrolling);
			break;
		case PROP_SCROLLING_DECELERATION:
			g_value_set_double(value, scrolling_actor->priv->scrolling_deceleration);
			break;
	        default:
		        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		        break;
        }
}

static void tangle_scrolling_actor_dispose(GObject* object) {
	TangleScrollingActor* scrolling_actor;
	
	scrolling_actor = TANGLE_SCROLLING_ACTOR(object);
	
	if (scrolling_actor->priv->captured_event_handler_id) {
		g_signal_handler_disconnect(scrolling_actor, scrolling_actor->priv->captured_event_handler_id);
		scrolling_actor->priv->captured_event_handler_id = 0;
	}

	G_OBJECT_CLASS(tangle_scrolling_actor_parent_class)->dispose(object);
}

static void tangle_scrolling_actor_class_init(TangleScrollingActorClass* scrolling_actor_class) {
	GObjectClass* gobject_class = G_OBJECT_CLASS(scrolling_actor_class);
	ClutterActorClass* clutter_actor_class = CLUTTER_ACTOR_CLASS(scrolling_actor_class);
	TangleActorClass* actor_class = TANGLE_ACTOR_CLASS(scrolling_actor_class);
	TangleWrapperActorClass* wrapper_actor_class = TANGLE_WRAPPER_ACTOR_CLASS(scrolling_actor_class);

	gobject_class->dispose = tangle_scrolling_actor_dispose;
	gobject_class->set_property = tangle_scrolling_actor_set_property;
	gobject_class->get_property = tangle_scrolling_actor_get_property;
	
	clutter_actor_class->button_press_event = tangle_scrolling_actor_button_press_event;
	clutter_actor_class->map = tangle_scrolling_actor_map;
	
	wrapper_actor_class->allocate_wrapped = tangle_scrolling_actor_allocate_wrapped;
	wrapper_actor_class->paint_wrapped = tangle_scrolling_actor_paint_wrapped;
	
	scrolling_actor_class->clamp_scrolling_offset_x = tangle_scrolling_actor_clamp_scrolling_offset_x_impl;
	scrolling_actor_class->clamp_scrolling_offset_y = tangle_scrolling_actor_clamp_scrolling_offset_y_impl;
	
	/**
	 * TangleScrollingActor:scrolling-offset-x:
	 *
	 * The offset to the left side of the wrapped actor's bounding box.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_OFFSET_X,
	                                g_param_spec_float("scrolling-offset-x",
	                                                    "Scrolling offset X",
	                                                    "The offset to the left side of the wrapped actor's bounding box",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-offset-y:
	 *
	 * The offset to the top side of the wrapped actor's bounding box.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_OFFSET_Y,
	                                g_param_spec_float("scrolling-offset-y",
	                                                    "Scrolling offset Y",
	                                                    "The offset to the top side of the wrapped actor's bounding box",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-width:
	 *
	 * The maximum value for the scrollng-offset-x: property.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_WIDTH,
	                                g_param_spec_float("scrolling-width",
	                                                    "Scrolling width",
	                                                    "The maximum value for the scrollng-offset-y property",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-height:
	 *
	 * The maximum value for the scrollng-offset-y: property.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_HEIGHT,
	                                g_param_spec_float("scrolling-height",
	                                                    "Scrolling height",
	                                                    "The maximum value for the scrollng-offset-y property",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-threshold-x:
	 *
	 * The horizontal threshold before start scrolling.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_THRESHOLD_X,
	                                g_param_spec_float("scrolling-threshold-x",
	                                                    "Scrolling threshold X",
	                                                    "The horizontal threshold before start scrolling",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-threshold-y:
	 *
	 * The vertical threshold before start scrolling.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_THRESHOLD_Y,
	                                g_param_spec_float("scrolling-threshold-y",
	                                                    "Scrolling threshold Y",
	                                                    "The vertical threshold before start scrolling",
	                                                   -G_MAXFLOAT, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:max-overshoot-x:
	 *
	 * The horizontal maximum of overshooting.
	 */
	g_object_class_install_property(gobject_class, PROP_MAX_OVERSHOOT_X,
	                                g_param_spec_float("max-overshoot-x",
	                                                    "Max overshoot X",
	                                                    "The horizontal maximum of overshooting",
	                                                   0.0, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:max-overshoot-y:
	 *
	 * The vertical maximum of overshooting.
	 */
	g_object_class_install_property(gobject_class, PROP_MAX_OVERSHOOT_Y,
	                                g_param_spec_float("max-overshoot-y",
	                                                    "Max overshoot Y",
	                                                    "The vertical maximum of overshooting",
	                                                   0.0, G_MAXFLOAT, 0.0,
	                                                   G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:repeat-x:
	 *
	 * Whether to repeat the wrapped actor painting when overshooting horizontally.
	 * Setting this to FALSE sets also :continuous-x to FALSE as a side effect.
	 */
	g_object_class_install_property(gobject_class, PROP_REPEAT_X,
	                                g_param_spec_boolean("repeat-x",
	                                                    "Repeat X",
	                                                    "Whether to repeat the wrapped actor painting when overshooting horizontally",
	                                                    FALSE,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:repeat-y:
	 *
	 * Whether to repeat the wrapped actor painting when overshooting vertically.
	 * Setting this to FALSE sets also :continuous-y to FALSE as a side effect.
	 */
	g_object_class_install_property(gobject_class, PROP_REPEAT_Y,
	                                g_param_spec_boolean("repeat-y",
	                                                    "Repeat Y",
	                                                    "Whether to repeat the wrapped actor painting when overshooting vertically",
	                                                    FALSE,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:continuous-x:
	 *
	 * Whether to connect the left and right edges of the wrapped actor to enable infinite horizontal scrolling.
	 * Setting this to TRUE sets also :repeat-x to TRUE as a side effect.
	 */
	g_object_class_install_property(gobject_class, PROP_CONTINUOUS_X,
	                                g_param_spec_boolean("continuous-x",
	                                                    "Continuous X",
	                                                    "Whether to connect the left and right edges of the wrapped actor to enable infinite horizontal scrolling",
	                                                    FALSE,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:continuous-y:
	 *
	 * Whether to connect the top and bottom edges of the wrapped actor to enable infinite vertical scrolling.
	 * Setting this to TRUE sets also :repeat-y to TRUE as a side effect.
	 */
	g_object_class_install_property(gobject_class, PROP_CONTINUOUS_Y,
	                                g_param_spec_boolean("continuous-y",
	                                                    "Continuous Y",
	                                                    "Whether to connect the top and bottom edges of the wrapped actor to enable infinite vertical scrolling",
	                                                    FALSE,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:kinetic-scrolling:
	 *
	 * Whether kinetic scrolling is enabled or not.
	 */
	g_object_class_install_property(gobject_class, PROP_KINETIC_SCROLLING,
	                                g_param_spec_boolean("kinetic-scrolling",
	                                                    "Kinetic scrolling",
	                                                    "Whether kinetic scrolling is enabled or not",
	                                                    TRUE,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
	/**
	 * TangleScrollingActor:scrolling-deceleration:
	 *
	 * Deceleration of kinetic scrolling.
	 */
	g_object_class_install_property(gobject_class, PROP_SCROLLING_DECELERATION,
	                                g_param_spec_boolean("scrolling-deceleration",
	                                                    "Scrolling deceleration",
	                                                    "Deceleration of kinetic scrolling",
	                                                    0.7,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));

	signals[CLAMP_SCROLLING_OFFSET_X] = g_signal_new("clamp-scrolling-offset-x",
	                                       G_TYPE_FROM_CLASS(gobject_class),
					       G_SIGNAL_RUN_LAST,
					       G_STRUCT_OFFSET(TangleScrollingActorClass, clamp_scrolling_offset_x),
					       NULL, NULL,
					       g_cclosure_marshal_VOID__OBJECT,
					       G_TYPE_NONE, 1,
					       TANGLE_TYPE_CLAMP);
	signals[CLAMP_SCROLLING_OFFSET_Y] = g_signal_new("clamp-scrolling-offset-y",
	                                       G_TYPE_FROM_CLASS(gobject_class),
					       G_SIGNAL_RUN_LAST,
					       G_STRUCT_OFFSET(TangleScrollingActorClass, clamp_scrolling_offset_y),
					       NULL, NULL,
					       g_cclosure_marshal_VOID__OBJECT,
					       G_TYPE_NONE, 1,
					       TANGLE_TYPE_CLAMP);

	g_type_class_add_private(gobject_class, sizeof(TangleScrollingActorPrivate));
}

static void tangle_scrolling_actor_init(TangleScrollingActor* scrolling_actor) {
	scrolling_actor->priv = G_TYPE_INSTANCE_GET_PRIVATE(scrolling_actor, TANGLE_TYPE_SCROLLING_ACTOR, TangleScrollingActorPrivate);
	scrolling_actor->priv->kinetic_scrolling = TRUE;
	scrolling_actor->priv->scrolling_deceleration = 0.7;

	clutter_actor_set_reactive(CLUTTER_ACTOR(scrolling_actor), TRUE);
	g_object_set(CLUTTER_ACTOR(scrolling_actor), "clip-to-allocation", TRUE, NULL);
}

static void on_overshooting_animation_completed(ClutterAnimation* animation, gpointer user_data) {
	TangleVault* vault;
	TangleScrollingActor* scrolling_actor;
	const gchar* property;
	gfloat value;
	
	vault = TANGLE_VAULT(user_data);
	tangle_vault_get(vault, 3, TANGLE_TYPE_SCROLLING_ACTOR, &scrolling_actor, G_TYPE_STRING, &property, G_TYPE_FLOAT, &value);
	
	tangle_actor_animate(TANGLE_ACTOR(scrolling_actor), CLUTTER_EASE_IN_OUT_QUAD, 350, property, value, NULL);
}

static void handle_overshooting_animation(TangleScrollingActor* scrolling_actor, ClutterAnimation* animation, const gchar* property, gfloat value) {
	TangleVault* vault;

	vault = tangle_vault_new(3, TANGLE_TYPE_SCROLLING_ACTOR, scrolling_actor, G_TYPE_STRING, property, G_TYPE_FLOAT, value);
	tangle_signal_connect_vault_flags(animation, "completed", G_CALLBACK(on_overshooting_animation_completed), vault, G_CONNECT_AFTER);
}

#define ABS_F(x) ((x) < 0 ? -(x) : (x))

static void handle_event(TangleScrollingActor* scrolling_actor, Axis axis, gfloat coordinate, gboolean release, guint event_time) {
	ClutterBackend* backend;
	gfloat d;
	gdouble v, a;
	guint t;
	TangleClamp* clamp;
	gfloat delta = 0.0;

	backend = clutter_get_default_backend();
	d = coordinate - scrolling_actor->priv->start_scrolling_motion[axis];

	if (scrolling_actor->priv->scrolling_effective[axis] ||
	    (!scrolling_actor->priv->descendant_dragging &&
	     (ABS_F(d) > scrolling_actor->priv->scrolling_threshold[axis]))) {

		if (!scrolling_actor->priv->scrolling_effective[axis] ||
		    event_time - scrolling_actor->priv->previous_motion_time > clutter_backend_get_double_click_time(backend)) {
			if (!scrolling_actor->priv->scrolling_effective[axis]) {		
				scrolling_actor->priv->scrolling_effective[axis] = TRUE;
				scrolling_actor->priv->start_scrolling_motion[axis] = coordinate;
				scrolling_actor->priv->start_scrolling_offset[axis] = scrolling_actor->priv->scrolling_offset[axis];
				d = 0;
			}
			scrolling_actor->priv->going_forward[axis] = scrolling_actor->priv->start_motion[axis] < coordinate;
			scrolling_actor->priv->start_motion[axis] = scrolling_actor->priv->farest_motion[axis] = coordinate;
			scrolling_actor->priv->start_motion_time[axis] = scrolling_actor->priv->farest_motion_time[axis] = event_time;

		} else {
			if ((scrolling_actor->priv->going_forward[axis] && coordinate > scrolling_actor->priv->farest_motion[axis]) ||
			    (!scrolling_actor->priv->going_forward[axis] && coordinate < scrolling_actor->priv->farest_motion[axis])) {
				scrolling_actor->priv->farest_motion[axis] = coordinate;
				scrolling_actor->priv->farest_motion_time[axis] = event_time;
			} else if (ABS_F(coordinate - scrolling_actor->priv->farest_motion[axis]) > clutter_backend_get_double_click_distance(backend)) {				
				scrolling_actor->priv->start_motion[axis] = scrolling_actor->priv->farest_motion[axis];
				scrolling_actor->priv->start_motion_time[axis] = scrolling_actor->priv->farest_motion_time[axis];
				scrolling_actor->priv->farest_motion[axis] = coordinate;
				scrolling_actor->priv->farest_motion_time[axis] = event_time;
				scrolling_actor->priv->going_forward[axis] = scrolling_actor->priv->start_motion[axis] < coordinate;
			}
		}

		if (axis == HORIZONTAL) {
			tangle_scrolling_actor_set_scrolling_offset_x(scrolling_actor, scrolling_actor->priv->start_scrolling_offset[axis] - d);
		} else {
			tangle_scrolling_actor_set_scrolling_offset_y(scrolling_actor, scrolling_actor->priv->start_scrolling_offset[axis] - d);
		}
		
		if (release && scrolling_actor->priv->kinetic_scrolling &&
		    event_time - scrolling_actor->priv->previous_motion_time < clutter_backend_get_double_click_time(backend)) {
			if (event_time - scrolling_actor->priv->start_motion_time[axis] > 0) {
				v = (gdouble)(coordinate - scrolling_actor->priv->start_motion[axis]) / (event_time - scrolling_actor->priv->start_motion_time[axis]);

				clamp = tangle_clamp_new((axis == HORIZONTAL ? TANGLE_CLAMP_HORIZONTAL : TANGLE_CLAMP_VERTICAL), normalise_scrolling_offset(scrolling_actor, axis, scrolling_actor->priv->scrolling_offset[axis] + (scrolling_actor->priv->going_forward[axis] ? -1 : 1) * v * v * 1000 / scrolling_actor->priv->scrolling_deceleration / 2, &delta));
				g_signal_emit(scrolling_actor, signals[(axis == HORIZONTAL ? CLAMP_SCROLLING_OFFSET_X : CLAMP_SCROLLING_OFFSET_Y)], 0, clamp);
				if (tangle_clamp_get_clamped_value_set(clamp)) {
					a = v * v * 1000 / ABS_F(tangle_clamp_get_clamped_value(clamp) - scrolling_actor->priv->scrolling_offset[axis] + delta) / 2;
				} else {
					a = scrolling_actor->priv->scrolling_deceleration;

				}
				g_object_unref(clamp);

				t = ABS_F(v * 1000.0 / a);

				if (t > 0) {
					d = (scrolling_actor->priv->going_forward[axis] ? -1 : 1) * v * v * 1000 / a / 2;
					if (scrolling_actor->priv->continuous[axis]) {
						scrolling_actor->priv->animation[axis] = tangle_actor_animate(TANGLE_ACTOR(scrolling_actor), CLUTTER_EASE_OUT_QUAD, t,
													           (axis == HORIZONTAL ? "scrolling-offset-x" : "scrolling-offset-y"), scrolling_actor->priv->scrolling_offset[axis] + d,
													           NULL);
						g_object_add_weak_pointer(G_OBJECT(scrolling_actor->priv->animation[axis]), (gpointer*)&scrolling_actor->priv->animation[axis]);
						g_signal_connect_swapped(scrolling_actor->priv->animation[axis], "completed", G_CALLBACK((axis == HORIZONTAL ? tangle_scrolling_actor_clamp_scrolling_offset_x : tangle_scrolling_actor_clamp_scrolling_offset_y)), scrolling_actor);
					} else {
						if (scrolling_actor->priv->scrolling_offset[axis] + d < -scrolling_actor->priv->max_overshoot[axis]) {
							t *= -(scrolling_actor->priv->scrolling_offset[axis] + scrolling_actor->priv->max_overshoot[axis]) / d;
							d = -(scrolling_actor->priv->scrolling_offset[axis] + scrolling_actor->priv->max_overshoot[axis]);
						} else if (scrolling_actor->priv->scrolling_offset[axis] + d > scrolling_actor->priv->scrolling_length[axis] + scrolling_actor->priv->max_overshoot[axis]) {
							t *= (scrolling_actor->priv->scrolling_length[axis] - scrolling_actor->priv->scrolling_offset[axis] + scrolling_actor->priv->max_overshoot[axis]) / d;
							d = (scrolling_actor->priv->scrolling_length[axis] - scrolling_actor->priv->scrolling_offset[axis] + scrolling_actor->priv->max_overshoot[axis]);					
						}

						if (t > 0) {
							scrolling_actor->priv->animation[axis] = tangle_actor_animate(TANGLE_ACTOR(scrolling_actor), CLUTTER_EASE_OUT_QUAD, t,
														  (axis == HORIZONTAL ? "scrolling-offset-x" : "scrolling-offset-y"), scrolling_actor->priv->scrolling_offset[axis] + d,
														  NULL);
							g_object_add_weak_pointer(G_OBJECT(scrolling_actor->priv->animation[axis]), (gpointer*)&scrolling_actor->priv->animation[axis]);
							g_signal_connect_swapped(scrolling_actor->priv->animation[axis], "completed", G_CALLBACK((axis == HORIZONTAL ? tangle_scrolling_actor_clamp_scrolling_offset_x : tangle_scrolling_actor_clamp_scrolling_offset_y)), scrolling_actor);
							if (scrolling_actor->priv->scrolling_offset[axis] + d < 0.0) {
								handle_overshooting_animation(scrolling_actor, scrolling_actor->priv->animation[axis], (axis == HORIZONTAL ? "scrolling-offset-x" : "scrolling-offset-y"), 0.0);
							} else if (scrolling_actor->priv->scrolling_offset[axis] + d > scrolling_actor->priv->scrolling_length[axis]) {
								handle_overshooting_animation(scrolling_actor, scrolling_actor->priv->animation[axis], (axis == HORIZONTAL ? "scrolling-offset-x" : "scrolling-offset-y"), scrolling_actor->priv->scrolling_length[axis]);
							}					
						} else if (axis == HORIZONTAL) {
							tangle_scrolling_actor_clamp_scrolling_offset_x(scrolling_actor);
						} else {
							tangle_scrolling_actor_clamp_scrolling_offset_y(scrolling_actor);
						}
					}
				} else if (axis == HORIZONTAL) {
					tangle_scrolling_actor_clamp_scrolling_offset_x(scrolling_actor);
				} else {
					tangle_scrolling_actor_clamp_scrolling_offset_y(scrolling_actor);
				}
			}
		} else if (release) {
			tangle_scrolling_actor_clamp_offsets(scrolling_actor);
		}

		scrolling_actor->priv->previous_motion_time = event_time;
	} else if (release) {
		if (axis == HORIZONTAL) {
			tangle_scrolling_actor_clamp_scrolling_offset_x(scrolling_actor);
		} else {
			tangle_scrolling_actor_clamp_scrolling_offset_y(scrolling_actor);
		}	
	}
}

static gboolean on_captured_event(ClutterActor* actor, ClutterEvent* event, gpointer user_data) {
	TangleScrollingActor* scrolling_actor;
	gfloat x, y, dx, dy;
	gdouble vx, vy, ax, ay;
	guint tx, ty;
	TangleClamp* clamp;
	gfloat delta;

	if (event->type == CLUTTER_MOTION || event->type == CLUTTER_BUTTON_RELEASE) {
		scrolling_actor = TANGLE_SCROLLING_ACTOR(user_data);	
		
		g_object_ref(scrolling_actor);
		
		clutter_event_get_coords(event, &x, &y);
		
		handle_event(scrolling_actor, HORIZONTAL, x, event->type == CLUTTER_BUTTON_RELEASE, clutter_event_get_time(event));
		handle_event(scrolling_actor, VERTICAL, y, event->type == CLUTTER_BUTTON_RELEASE, clutter_event_get_time(event));

		if (event->type == CLUTTER_BUTTON_RELEASE) {
			stop_interacting(scrolling_actor);
		}

		g_object_unref(scrolling_actor);
	}

	return FALSE;
}

static gfloat get_length(TangleScrollingActor* scrolling_actor, Axis axis) {
	ClutterActorBox actor_box;
	gfloat length;
	
	clutter_actor_get_allocation_box(CLUTTER_ACTOR(scrolling_actor), &actor_box);
	if (axis == HORIZONTAL) {
		length = actor_box.x2 - actor_box.x1;
	} else {
		length = actor_box.y2 - actor_box.y1;
	}
	if (length == 0.0) {
		length = G_MINFLOAT;
	}
	
	return length;
}

static gfloat normalise_scrolling_offset(TangleScrollingActor* scrolling_actor, Axis axis, gfloat scrolling_offset, gfloat* delta_return) {
	gfloat length;
	
	if (scrolling_actor->priv->continuous[axis]) {
		length = scrolling_actor->priv->scrolling_length[axis] + get_length(scrolling_actor, axis);
		if (scrolling_offset < 0.0) {
			scrolling_offset -= (((gint)(scrolling_offset / length)) - 1) * length;
			if (delta_return) {
				*delta_return = (((gint)(scrolling_offset / length)) - 1) * length;
			}
		} else if (scrolling_offset > scrolling_actor->priv->scrolling_length[axis] + length) {
			scrolling_offset -= (((gint)(scrolling_offset / length))) * length;
			if (delta_return) {
				*delta_return = (((gint)(scrolling_offset / length))) * length;
			}
		}
	} else {
		if (scrolling_offset < -scrolling_actor->priv->max_overshoot[axis]) {
			if (delta_return) {
				*delta_return = -scrolling_actor->priv->max_overshoot[axis] - scrolling_offset;
			}
			scrolling_offset = -scrolling_actor->priv->max_overshoot[axis];
		} else if (scrolling_offset > scrolling_actor->priv->scrolling_length[axis] + scrolling_actor->priv->max_overshoot[axis]) {
			if (delta_return) {
				*delta_return = scrolling_actor->priv->scrolling_length[axis] + scrolling_actor->priv->max_overshoot[axis] - scrolling_offset;
			}
			scrolling_offset = scrolling_actor->priv->scrolling_length[axis] + scrolling_actor->priv->max_overshoot[axis];
		}
	}
	
	return scrolling_offset;
}

static void stop_interacting(TangleScrollingActor* scrolling_actor) {
	if (scrolling_actor->priv->captured_event_handler_id) {
		g_signal_handler_disconnect(clutter_actor_get_stage(CLUTTER_ACTOR(scrolling_actor)), scrolling_actor->priv->captured_event_handler_id);
		scrolling_actor->priv->captured_event_handler_id = 0;
	}
}

static void create_clone(ClutterActor** actor_p, ClutterActor* wrapped_actor, ClutterActor* parent_actor) {
	if (!*actor_p) {
		*actor_p = tangle_clone_new(wrapped_actor);
		clutter_actor_set_parent(*actor_p, parent_actor);
	}
}

static void destroy_clone(ClutterActor** actor_p) {
	if (*actor_p) {
		clutter_actor_unparent(*actor_p);
		*actor_p = NULL;
	}
}

static void setup_clones(TangleScrollingActor* scrolling_actor) {
	ClutterActor* wrapped_actor;
	ClutterActor* parent_actor;
	
	clutter_actor_push_internal(CLUTTER_ACTOR(scrolling_actor));
	
	wrapped_actor = tangle_wrapper_actor_get_wrapped(TANGLE_WRAPPER_ACTOR(scrolling_actor));
	parent_actor = CLUTTER_ACTOR(scrolling_actor);
	
	if (scrolling_actor->priv->repeat[VERTICAL]) {
		create_clone(&scrolling_actor->priv->clones[1], wrapped_actor, parent_actor);
		create_clone(&scrolling_actor->priv->clones[6], wrapped_actor, parent_actor);
	} else {
		destroy_clone(&scrolling_actor->priv->clones[0]);
		destroy_clone(&scrolling_actor->priv->clones[1]);
		destroy_clone(&scrolling_actor->priv->clones[2]);
		destroy_clone(&scrolling_actor->priv->clones[5]);
		destroy_clone(&scrolling_actor->priv->clones[6]);
		destroy_clone(&scrolling_actor->priv->clones[7]);
	}
	
	if (scrolling_actor->priv->repeat[HORIZONTAL]) {
		create_clone(&scrolling_actor->priv->clones[3], wrapped_actor, parent_actor);
		create_clone(&scrolling_actor->priv->clones[4], wrapped_actor, parent_actor);
	} else {
		destroy_clone(&scrolling_actor->priv->clones[0]);
		destroy_clone(&scrolling_actor->priv->clones[2]);
		destroy_clone(&scrolling_actor->priv->clones[3]);
		destroy_clone(&scrolling_actor->priv->clones[4]);
		destroy_clone(&scrolling_actor->priv->clones[5]);
		destroy_clone(&scrolling_actor->priv->clones[7]);
	}
	
	if (scrolling_actor->priv->repeat[VERTICAL] && scrolling_actor->priv->repeat[HORIZONTAL]) {
		create_clone(&scrolling_actor->priv->clones[0], wrapped_actor, parent_actor);
		create_clone(&scrolling_actor->priv->clones[2], wrapped_actor, parent_actor);
		create_clone(&scrolling_actor->priv->clones[5], wrapped_actor, parent_actor);
		create_clone(&scrolling_actor->priv->clones[7], wrapped_actor, parent_actor);	
	}
	
	clutter_actor_pop_internal(CLUTTER_ACTOR(scrolling_actor));	
}
