/*
 * tangle-wrapper-actor.h
 *
 * This file is part of Tangle Toolkit - A graphical actor library based on Clutter Toolkit
 *
 * (c) 2010 Henrik Hedberg <henrik.hedberg@innologies.fi>
 *
 */

#include "tangle-wrapper-actor.h"
#include <string.h>
#include <gobject/gvaluecollector.h>

/**
 * SECTION:tangle-wrapper-actor
 * @Short_description: 	A #TangleActor that wraps any ClutterActor inside
 * @Title: TangleWrapperActor
 */

static void clutter_container_iface_init(ClutterContainerIface *iface);
static void clutter_scriptable_iface_init(ClutterScriptableIface *iface);

G_DEFINE_TYPE_WITH_CODE(TangleWrapperActor, tangle_wrapper_actor, TANGLE_TYPE_ACTOR,
                        G_IMPLEMENT_INTERFACE(CLUTTER_TYPE_CONTAINER, clutter_container_iface_init);
                        G_IMPLEMENT_INTERFACE(CLUTTER_TYPE_SCRIPTABLE, clutter_scriptable_iface_init););

enum {
	PROP_0,
	PROP_WRAPPED
};

struct _TangleWrapperActorPrivate {
	ClutterActor* wrapped;
};

static ClutterScriptableIface* parent_scriptable_iface = NULL;

ClutterActor* tangle_wrapper_actor_new(ClutterActor* wrapped) {

	return CLUTTER_ACTOR(g_object_new(TANGLE_TYPE_WRAPPER_ACTOR, "wrapped", wrapped, NULL));
}

ClutterActor* tangle_wrapper_actor_get_wrapped(TangleWrapperActor* wrapper_actor) {

	return wrapper_actor->priv->wrapped;
}

void tangle_wrapper_actor_get(TangleWrapperActor* wrapper_actor, const gchar* first_property_name, ...) {
	const gchar* property_name;
	va_list args;
	GParamSpec* param_spec;
	GObject* object;
	GValue value;
	gchar* error;
	
	va_start(args, first_property_name);
	for (property_name = first_property_name; property_name; property_name = va_arg(args, const gchar*)) {
		if ((param_spec = g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor)), property_name))) {
			object = G_OBJECT(wrapper_actor);
		} else if ((param_spec = g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor->priv->wrapped)), property_name))) {
			object = G_OBJECT(wrapper_actor->priv->wrapped);
		} else {
			g_warning("%s: Neither wrapper nor wrapped actor has property named '%s'.", G_STRLOC, property_name);
			break;
		}

		g_object_get_property(object, property_name, &value);

		error = NULL;
		G_VALUE_LCOPY(&value, args, 0, &error);

		g_value_unset(&value);

		if (error) {
			g_warning("%s: %s", G_STRLOC, error);
			g_free(error);			
			break;
		}
	}
	va_end(args);	
}

void tangle_wrapper_actor_set(TangleWrapperActor* wrapper_actor, const gchar* first_property_name, ...) {
	const gchar* property_name;
	va_list args;
	GParamSpec* param_spec;
	GObject* object;
	GValue value;
	gchar* error;
	
	va_start(args, first_property_name);
	for (property_name = first_property_name; property_name; property_name = va_arg(args, const gchar*)) {
		if ((param_spec = g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor)), property_name))) {
			object = G_OBJECT(wrapper_actor);
		} else if ((param_spec = g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor->priv->wrapped)), property_name))) {
			object = G_OBJECT(wrapper_actor->priv->wrapped);
		} else {
			g_warning("%s: Neither wrapper nor wrapped actor has property named '%s'.", G_STRLOC, property_name);
			break;
		}

		g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(param_spec));		
		error = NULL;
		G_VALUE_COLLECT(&value, args, 0, &error);
		if (error) {
			g_warning("%s: %s", G_STRLOC, error);
			g_free(error);
			break;
		}
		
		g_object_set_property(object, property_name, &value);

		g_value_unset(&value);
	}
	va_end(args);
}

static void tangle_wrapper_actor_get_preferred_width(TangleActor* actor, gfloat for_height, gboolean interacting, gfloat* min_width_p, gfloat* natural_width_p, gfloat* max_width_p) {
	TangleWrapperActor* wrapper_actor;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(actor);
	if (TANGLE_IS_ACTOR(wrapper_actor->priv->wrapped)) {
		tangle_actor_get_preferred_width(TANGLE_ACTOR(wrapper_actor->priv->wrapped), for_height, interacting, min_width_p, natural_width_p, max_width_p);
	} else {
		clutter_actor_get_preferred_width(wrapper_actor->priv->wrapped, for_height, min_width_p, natural_width_p);
		if (max_width_p) {
			*max_width_p = 0;
		}
	}
}

static void tangle_wrapper_actor_get_preferred_height(TangleActor* actor, gfloat for_width, gboolean interacting, gfloat* min_height_p, gfloat* natural_height_p, gfloat* max_height_p) {
	TangleWrapperActor* wrapper_actor;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(actor);
	if (TANGLE_IS_ACTOR(wrapper_actor->priv->wrapped)) {
		tangle_actor_get_preferred_height(TANGLE_ACTOR(wrapper_actor->priv->wrapped), for_width, interacting, min_height_p, natural_height_p, max_height_p);
	} else {
		clutter_actor_get_preferred_height(TANGLE_WRAPPER_ACTOR(actor)->priv->wrapped, for_width, min_height_p, natural_height_p);
		if (max_height_p) {
			*max_height_p = 0;
		}
	}
}

static void tangle_wrapper_actor_allocate_wrapped(TangleWrapperActor* wrapper_actor, ClutterActor* wrapped, const ClutterActorBox* box, ClutterAllocationFlags flags) {
	clutter_actor_allocate(wrapped, box, flags);
}

static void tangle_wrapper_actor_allocate(ClutterActor* actor, const ClutterActorBox* box, ClutterAllocationFlags flags) {
	TangleWrapperActor* wrapper_actor;
	ClutterActorBox actor_box;
	TangleWrapperActorClass* wrapper_actor_class;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(actor);
	
	CLUTTER_ACTOR_CLASS(tangle_wrapper_actor_parent_class)->allocate(actor, box, flags);

	actor_box.x1 = actor_box.y1 = 0.0;
	actor_box.x2 = box->x2 - box->x1;
	actor_box.y2 = box->y2 - box->y1;
	
	TANGLE_WRAPPER_ACTOR_GET_CLASS(wrapper_actor)->allocate_wrapped(wrapper_actor, wrapper_actor->priv->wrapped, &actor_box, flags);
}

static void tangle_wrapper_actor_paint_wrapped(TangleWrapperActor* wrapper_actor, ClutterActor* wrapped) {
		clutter_actor_paint(wrapper_actor->priv->wrapped);
}

static void tangle_wrapper_actor_paint(ClutterActor* actor) {
	TangleWrapperActor* wrapper_actor;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(actor);

	if (CLUTTER_ACTOR_IS_MAPPED(wrapper_actor->priv->wrapped)) {
		cogl_push_matrix();

		tangle_actor_apply_transition_scale(TANGLE_ACTOR(wrapper_actor));

		TANGLE_WRAPPER_ACTOR_GET_CLASS(wrapper_actor)->paint_wrapped(wrapper_actor, wrapper_actor->priv->wrapped);

		cogl_pop_matrix();
	}
}

static void tangle_wrapper_actor_pick(ClutterActor* actor, const ClutterColor* color) {
	CLUTTER_ACTOR_CLASS(tangle_wrapper_actor_parent_class)->pick(actor, color);
	
	tangle_wrapper_actor_paint(actor);
}

static void tangle_wrapper_actor_map(ClutterActor* actor) {
	CLUTTER_ACTOR_CLASS(tangle_wrapper_actor_parent_class)->map(actor);

	g_return_if_fail(TANGLE_WRAPPER_ACTOR(actor)->priv->wrapped);

	clutter_actor_map(TANGLE_WRAPPER_ACTOR(actor)->priv->wrapped);
}

static void tangle_wrapper_actor_unmap(ClutterActor* actor) {
	CLUTTER_ACTOR_CLASS(tangle_wrapper_actor_parent_class)->unmap(actor);

	g_return_if_fail(TANGLE_WRAPPER_ACTOR(actor)->priv->wrapped);

	clutter_actor_unmap(TANGLE_WRAPPER_ACTOR(actor)->priv->wrapped);
}

static void tangle_wrapper_actor_set_property(GObject* object, guint prop_id, const GValue* value, GParamSpec* pspec) {
	TangleWrapperActor* wrapper_actor;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(object);

	switch (prop_id) {
		case PROP_WRAPPED:
			wrapper_actor->priv->wrapped = CLUTTER_ACTOR(g_value_get_object(value));
			clutter_actor_push_internal(CLUTTER_ACTOR(wrapper_actor));
			clutter_actor_set_parent(wrapper_actor->priv->wrapped, CLUTTER_ACTOR(wrapper_actor));
			clutter_actor_pop_internal(CLUTTER_ACTOR(wrapper_actor));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
			break;
	}
}

static void tangle_wrapper_actor_get_property(GObject* object, guint prop_id, GValue* value, GParamSpec* pspec) {
        TangleWrapperActor* wrapper_actor;

	wrapper_actor = TANGLE_WRAPPER_ACTOR(object);

        switch (prop_id) {
		case PROP_WRAPPED:
			g_value_set_object(value, wrapper_actor->priv->wrapped);
			break;
		default:
		        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		        break;
        }
}

static void tangle_wrapper_actor_dispose(GObject* object) {
	TangleWrapperActor* wrapper_actor;
	
	wrapper_actor = TANGLE_WRAPPER_ACTOR(object);
	
	if (wrapper_actor->priv->wrapped) {
		clutter_actor_unparent(wrapper_actor->priv->wrapped);
		wrapper_actor->priv->wrapped = NULL;
	}

	G_OBJECT_CLASS(tangle_wrapper_actor_parent_class)->dispose(object);
}

static void tangle_wrapper_actor_class_init(TangleWrapperActorClass* wrapper_actor_class) {
	GObjectClass* gobject_class = G_OBJECT_CLASS(wrapper_actor_class);
	ClutterActorClass* clutter_actor_class = CLUTTER_ACTOR_CLASS(wrapper_actor_class);
	TangleActorClass* actor_class = TANGLE_ACTOR_CLASS(wrapper_actor_class);

	gobject_class->dispose = tangle_wrapper_actor_dispose;
	gobject_class->set_property = tangle_wrapper_actor_set_property;
	gobject_class->get_property = tangle_wrapper_actor_get_property;

	clutter_actor_class->allocate = tangle_wrapper_actor_allocate;
	clutter_actor_class->paint = tangle_wrapper_actor_paint;
	clutter_actor_class->pick = tangle_wrapper_actor_pick;
	clutter_actor_class->map = tangle_wrapper_actor_map;
	clutter_actor_class->unmap = tangle_wrapper_actor_unmap;
	
	actor_class->get_preferred_width = tangle_wrapper_actor_get_preferred_width;
	actor_class->get_preferred_height = tangle_wrapper_actor_get_preferred_height;

	wrapper_actor_class->allocate_wrapped = tangle_wrapper_actor_allocate_wrapped;
	wrapper_actor_class->paint_wrapped = tangle_wrapper_actor_paint_wrapped;

	/**
	 * TangleWrapperActor:actor:
	 */
	g_object_class_install_property(gobject_class, PROP_WRAPPED,
	                                g_param_spec_object("wrapped",
	                                                    "Wrapped",
	                                                    "The wrapped ClutterActor",
	                                                    CLUTTER_TYPE_ACTOR,
	                                                    G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));

	g_type_class_add_private(gobject_class, sizeof(TangleWrapperActorPrivate));
}

static void tangle_wrapper_actor_init(TangleWrapperActor* wrapper_actor) {
	wrapper_actor->priv = G_TYPE_INSTANCE_GET_PRIVATE(wrapper_actor, TANGLE_TYPE_WRAPPER_ACTOR, TangleWrapperActorPrivate);
}

static void tangle_wrapper_actor_add(ClutterContainer* container, ClutterActor* actor) {	
	g_critical("TangleWrapperActor does not support clutter_container_add().");
}

static void tangle_wrapper_actor_remove(ClutterContainer* container, ClutterActor* actor) {
	g_critical("TangleWrapperActor does not support clutter_container_remove().");
}

static void tangle_wrapper_actor_foreach(ClutterContainer* container, ClutterCallback callback, gpointer user_data) {
	g_return_if_fail(TANGLE_IS_WRAPPER_ACTOR(container));

	callback(TANGLE_WRAPPER_ACTOR(container)->priv->wrapped, user_data);
}

static void tangle_wrapper_actor_raise(ClutterContainer* container, ClutterActor* actor, ClutterActor* sibling) {
	g_return_if_fail(TANGLE_IS_WRAPPER_ACTOR(container));
}

static void tangle_wrapper_actor_lower(ClutterContainer* container, ClutterActor* actor, ClutterActor* sibling) {
	g_return_if_fail(TANGLE_IS_WRAPPER_ACTOR(container));
}

static void tangle_wrapper_actor_sort_depth_order(ClutterContainer* container) {
	g_return_if_fail(TANGLE_IS_WRAPPER_ACTOR(container));
}

static void clutter_container_iface_init(ClutterContainerIface* iface) {
	iface->add = tangle_wrapper_actor_add;
	iface->remove = tangle_wrapper_actor_remove;
	iface->foreach = tangle_wrapper_actor_foreach;
	iface->raise = tangle_wrapper_actor_raise;
	iface->lower = tangle_wrapper_actor_lower;
	iface->sort_depth_order = tangle_wrapper_actor_sort_depth_order;
}

static void tangle_wrapper_actor_set_custom_property(ClutterScriptable* scriptable, ClutterScript* script,  const gchar* name, const GValue* value) {
	TangleWrapperActor* wrapper_actor;

	wrapper_actor = TANGLE_WRAPPER_ACTOR(scriptable);

	if (!g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor)), name) &&
	    g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(wrapper_actor->priv->wrapped)), name)) {
		g_object_set_property( G_OBJECT(wrapper_actor->priv->wrapped), name, value);
	} else if (parent_scriptable_iface->set_custom_property) {
		parent_scriptable_iface->set_custom_property(scriptable, script, name, value);
	} else {
		g_object_set_property(G_OBJECT(scriptable), name, value);
	}
}

static void clutter_scriptable_iface_init(ClutterScriptableIface *iface) {
	if (!(parent_scriptable_iface = parent_scriptable_iface = g_type_interface_peek_parent (iface))) {
		parent_scriptable_iface = g_type_default_interface_peek(CLUTTER_TYPE_SCRIPTABLE);
	}

	iface->set_custom_property = tangle_wrapper_actor_set_custom_property;
}
